%------------------------------------------------------------------------------------
%
% A note on using this: you may want to change the following variables
%   noisome = true to get lots of verbiage about the solution process
%   plotit  = true to get a residual norm versus iteration number for the
%               iterative solve methods
%   droptol for when the ILUtp preconditioner is used (solver == 4)
% 
% Solves the linear system Ax = b, where A is a Matlab sparse matrix. The argument
% "solver" is an integer 1-2 indicating which solution method to use:
%
%   1. x = A\b;  [LU factorization]
%   2. x = bicgstab(A,b,tol,maxits); 
%   3. x = bicgstab with ILU(0) preconditioner
%   4. x = bicgstab with ILUp preconditioner
%
% The intent is to add more solve methods in the future. If the iterative solver
% bi-CGstab is used, this can also plot the convergence history (set plotit = true)
% and give some other info. Beware that bi-CGstab takes 2 matrix-vector products per
% iteration, and can stop after either one of those. Matlab returns that as a
% half-iteration, so it is possible for this to stop after 210.5 iterations. I plot
% the relative residual vector versus the number of matrix-vector products, so in
% case Matlab says 210.5 iterations, the plot goes to 421 instead. When noisome is
% true, this prints out the number of matrix-vector products, not the number of
% iterations to avoid the half-step count.
%
% Other issues: ILU(0) has no parameters to set (except a milu option; avoid it). But
% ILUtp has two, the more sensitive one is the droptol. I set it to a multiple (like
% 1.0e-2) of the smallest nonzero absolute value in the matrix A, but it may be
% better to do that for the largest nonzero absolute value. The storage cost can
% drastically increase as that value is decreased, so beware.
%
% The flag has value 0 if the solve succeeded.
%   flag = -1: unknown solver method specified
%   flag =  1: not enough input arguments given
%   flag =  2: bicgstab took maximum allowed number of iterations
%   flag =  3: bicgstab did not achieve required tolerance
%------------------------------------------------------------------------------------

function [x, flag] = solvit(A, b, solver, tol, maxits)

    noisome = true;     % Print out additional info
    plotit  = false;    % Plot the convergence for bi-CGstab

    flag = 0; % No errors so far

	if nargin < 3
		flag = -1;
		error('Need at least three arguments to solvit');
	end;

	if (solver == 1)
        if ~noisome
		    x    = A\b; 
        else
            [L,U] = lu(A);
            x = U\(L\b);
            storage = nnz(L) + nnz(U);
            Astore  = nnz(A);
            disp(sprintf('Storage for LU factors = %d nonzeros', storage))
            disp(sprintf('This is %g times as much as for A', storage/Astore))
        end
		return;

	elseif (solver >= 2)
		if nargin < 5
			flag = -3;
			error('Need at least five arguments to solvit for bicgstab');
		end;

        if solver == 4
            setup.type = 'ilutp';
            setup.droptol = (1.0e-2)*min(abs(A(A ~= 0)));
            setup.udiag  = 1;
            setup.thresh = 0.1;
            [L, U] = ilu(A, setup);
            if noisome
                precondnnz = nnz(L) + nnz(U);
                disp(sprintf('Number nonzeros from ilutp in solvit = %d', ...
                    precondnnz))
                disp(sprintf('So ilutp takes = %g as much memory as A',  ...
                    precondnnz/nnz(A)))
            end
		    [x,flag,relres,iter,resvec] = bicgstab(A,b,tol,maxits,L,U); 
        end

        if solver == 3
            setup.type = 'nofill';
            [L, U] = ilu(A, setup);
		    [x,flag,relres,iter,resvec] = bicgstab(A,b,tol,maxits,L,U); 
        end

        if solver == 2
		    [x,flag,relres,iter,resvec] = bicgstab(A,b,tol,maxits); 
        end

        if noisome
            disp(sprintf('bicgstab took %d iterations', 2*iter))
            disp(sprintf('final relative residual: %g ', relres))
        end

        if plotit
            figure;
            semilogy(1:length(resvec), resvec, 'b+-')
            xlabel('Number of Matrix-Vector Products')
            ylabel('Residual vector norm')
            grid on
            grid minor
            title('Convergence History for Bi-CGSTAb')
        end
        if (relres > tol)
            disp(sprintf('After bicgstab, relres = %g', relres))
            disp(sprintf('After bicgstab, tol/norm(b) = %g', tol/norm(b)))
            flag = 3;
            return
        end
		return;

        if (2*iter > maxits)
            flag = -2;
        end
		return;

	else
		disp('Unknown solver method in solvit');
        flag = -3;
	end

